# Copyright (c) 2015-2016, Workiva Inc.  All rights reserved.
# Copyright (c) 2015-2016, XBRL US Inc.  All rights reserved.
from collections import defaultdict
from arelle import ValidateXbrlCalcs


DEI_NAMESPACE_LIST = [
    'http://xbrl.sec.gov/dei/2014-01-31', 'http://xbrl.sec.gov/dei/2013-01-31',
    'http://xbrl.sec.gov/dei/2012-01-31', 'http://xbrl.sec.gov/dei/2011-01-31',
    'http://xbrl.us/dei/2009-01-31'
]

GAAP_NAMESPACE_LIST = [
    'http://fasb.org/us-gaap/2012-01-31', 'http://fasb.org/us-gaap/2013-01-31',
    'http://fasb.org/us-gaap/2014-01-31', 'http://fasb.org/us-gaap/2015-01-31'
]


class DecimalOverriddenFact(object):
    def __init__(self, fact, decimals_override):
        self.fact = fact
        self.decimals_override = decimals_override

    def get_decimals_override(self):
        return self.decimals_override

    def __getattr__(self, name):
        if name == 'decimals':
            return self.get_decimals_override()
        return getattr(self.fact, name)


def scale_values(facts):
    """
    Accepts a list of facts and returns a list of values from those facts
    re-scaled by the least precise fact's precision Makes no guarantees about
    data type, known types are integers, floats, and Decimals Precision
    attribute is not allowed in EFM 6.5.17, so this assumes decimals-use only.

    :param facts: list of facts to scale
    :type facts: list [:class:'~arelle.ModelInstanceObject.ModelFact']
    :return: list of values rescaled by least precise fact's precision
    :rtype: list [:class:'~decimals.Decimal']
    """
    decimals = set([fact.decimals for fact in facts])
    if len(decimals) == 1:
        # if all decimals match, then just return xValue
        # (their already scaled value)
        return [fact.xValue for fact in facts]
    # Redo the set getting only numbers
    decimals = {
        int(fact.decimals)
        for fact in facts
        if fact.decimals != 'INF' and
        fact.decimals is not None
    }
    min_decimals_str = str(min(decimals))
    return [
        ValidateXbrlCalcs.roundFact(
            DecimalOverriddenFact(fact, min_decimals_str),
            inferDecimals=True
        ) for fact in facts
    ]


def filter_duplicate_facts(facts, ignore_units=False):
    """
    This utility method should be used to prune duplicate facts from a set of
    facts.  The facts should all be the same concept, and the set returned will
    be all the unique facts in the set.

    :param facts: list of facts to find unique facts from
    :type facts: list ['~arelle.ModelInstanceObject.ModelFact']
    :param ignore_units: specifies whether units make fact unique or not
    :type ignore_units: bool
    :return: all unique facts in facts
    :rtype: list [:class:'~arelle.ModelInstanceObject.ModelFact']
    """
    mapped_facts = defaultdict(list)
    for f in facts:
        if f.contextID is not None and not f.isNil and f.xValid:
            mapped_facts[
                f.contextID, None if ignore_units else f.unitID
            ].append(f)
    result = []
    for v in mapped_facts.values():
        if len(v) == 1:
            result.extend(v)
    return result


def prepare_facts_for_calculation(fact_dict, unit_ignored_dict=None):
    """
    Takes a group of facts for calculations and groups them by context and
    units, and strips out the ones that
    are duplicated or not complete groupings.

    :param fact_dict: The fact dictionary.
        Should be {'conceptName':[facts_tagged_as_concept]...}
    :type fact_dict: dict
    :param unit_ignored_dict: The dictionary of concepts to ignore units for
        duplication checking. Should be {'conceptName': True...}. Defaults to
        'False' for anything not defined, and all units are tested otherwise.
    :type unit_ignored_dict: dict
    :return: A list of dicts that map a context-unit-matched set of the
        facts together.
        Should be: [{'conceptName1':fact, 'conceptName2':fact2,...}...]
    :rtype: list [dict]
    """
    unit_ignored = defaultdict(lambda: False)
    if unit_ignored_dict:
        unit_ignored.update(unit_ignored_dict)
    cleaned_fact_map = defaultdict()
    for k, v in fact_dict.items():
        cleaned_fact_map[k] = filter_duplicate_facts(
            v, ignore_units=unit_ignored[k]
        )
    result_map = defaultdict(dict)
    for k, v in cleaned_fact_map.items():
        if not unit_ignored[k]:
            for f in v:
                result_map[(f.contextID, f.unitID)][k] = f
    for k, v in cleaned_fact_map.items():
        if unit_ignored[k]:
            for f in v:
                for result_key, rm in result_map.items():
                    if result_key[0] == f.contextID:
                        rm[k] = f
    expected_len = len(fact_dict)
    return [m for m in result_map.values() if len(m) == expected_len]


def axis_exists(val, fact, axis_name):
    """
    Given a fact, check if fact is dimensionalized
    with given axis

    :param val: val with standard taxonomies dict to check
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param fact: fact to check
    :type fact: :class:'~arelle.ModelInstanceObject.ModelFact'
    :param axis_name: name of the axis to check
    :type axis_name: str
    :return: True if fact is demensionalized
    :rtype: bool
    """
    standard_taxonomies_dict = val.disclosureSystem.standardTaxonomiesDict

    return any(
        axis_name == dim.dimensionQname.localName
        for dim in fact.context.qnameDims.values()
        if dim.isExplicit and dim.dimensionQname and
        dim.dimensionQname.namespaceURI in standard_taxonomies_dict
    )


def member_exists(val, fact, member_name):
    """
    Given a fact, check if fact is dimensionalized with given axis

    :param val: val to check against
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param fact: fact to check against
    :type fact: :class:'~arelle.InstanceModelObject.ModelFact'
    :param member_name: member name to check against
    :type member_name: str
    :return: True if fact is demensionalized with gived axis
    :rtype: bool
    """

    standard_taxonomies_dict = val.disclosureSystem.standardTaxonomiesDict

    return any(
        member_name == dim.member.qname.localName
        for dim in fact.context.segDimValues.values()
        if dim.isExplicit and dim.member is not None and
        dim.member.qname.namespaceURI in standard_taxonomies_dict
    )


def axis_member_exists(val, fact, axis_name, member_name):
    """
    Given a fact, check if the fact is dimensionalized for a axis/member
    pairing

    :param val: val to check standard taxonomies dict on
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param fact: fact to check segDimValues of
    :type fact: :class:'~arelle.InstanceModelObject.ModelFact'
    :param axis_name: axis name to check against local name
    :type axis_name: str
    :param member_name: member_name to check against local name
    :type member_name: str
    :return: returns true if fact is dimensionalized
    :rtype: bool
    """
    standard_taxonomies_dict = val.disclosureSystem.standardTaxonomiesDict

    return any(
        member_name == dim.memberQname.localName and
        axis_name == dim.dimensionQname.localName
        for dim in fact.context.segDimValues.values()
        if dim.isExplicit and dim.memberQname is not None and
        dim.memberQname.namespaceURI in standard_taxonomies_dict and
        dim.dimensionQname is not None and
        dim.dimensionQname.namespaceURI in standard_taxonomies_dict
    )


def get_facts_with_type(lookup_type_strings, model_xbrl):
    """
    Returns a list of facts from the modelXbrl whose types match the types
    supplied in the lookup_type_strings list

    :param lookup_type_strings: string specifying what to lookup
    :type lookup_type_strings: str
    :param model_xbrl: modelXbrl to return facts from
    :type model_xbrl: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: list of facts from the specified modelXbrl
    :rtype: list [:class:'~arelle.ModelInstanceObject.ModelFact"]
    """
    list_types = []
    if lookup_type_strings:
        for type in lookup_type_strings:
            list_types.extend(
                f for f in list(model_xbrl.facts)
                if f.context is not None and
                f.concept.type is not None and f.concept.type.name == type
            )
    return list_types


def lookup_gaap_facts(fact_name, model_xbrl):
    """
    Returns the set of us-gaap facts for the
    given fact name.  Note that the facts will not be
    returned if the context or xValue is None.

    :param fact_name: name of the fact to lookup
    :type fact_name: str
    :param model_xbrl: modelXbrl to return facts from
    :type model_xbrl: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: list of us-gaap facts
    :rtype: list [:class:'~arelle.InstanceModelObject.ModelFact']
    """
    def valid_fact(fact):
        return (
            fact.concept.qname.namespaceURI in GAAP_NAMESPACE_LIST and
            fact.context is not None and fact.xValue is not None
        )
    facts = [
        f for f in model_xbrl.facts
        if f.concept.qname.localName == fact_name and valid_fact(f)
    ]
    return facts


def get_facts_dei(lookup_concept_strings, model_xbrl):
    """
    Returns a list of dei facts from the modelXbrl whose name matches the names
    supplied in the lookup_concept_strings list

    :param lookup_concept_strings: strings to look up using a loop
    :type lookup_concept_strings: list [str]
    :param model_xbrl: modelXbrl to get facts from
    :type model_xbrl: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: list of dei facts from specified model_xbrl
    :rtype: list [:class:'~arelle.InstanceModelObject.ModelFact']
    """
    list_dei = []
    if lookup_concept_strings:
        for dei in lookup_concept_strings:
            list_dei.extend(lookup_dei_facts(dei, model_xbrl, False))
    return list_dei


def lookup_dei_facts(fact_name, model_xbrl, validation=True):
    """
    Returns the set of dei facts for the
    given fact name

    :param fact_name: name of the fact
    :type fact_name: str
    :param model_xbrl: ModelXbrl to get the facts from
    :type model_xbrl: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param validation: should check for valid facts
    :type validation: bool
    :return: Set of dei facts for a given fact name
    :rtype: list [:class:'~arelle.InstanceModelObject.ModelFact']
    """
    facts = [
        f for f in model_xbrl.facts
        if f.concept.qname.localName == fact_name and
        f.concept.qname.namespaceURI in DEI_NAMESPACE_LIST
    ]
    if validation:
        facts = [
            f for f in facts
            if f.context is not None and f.xValue is not None
        ]
    return facts


LEGALENTITYAXIS_DEFAULT = ''


def legal_entity_axis_facts_by_member(facts):
    """
    Returns a dictionary of lists of facts, keyed off of the LegalEntityAxis
    member, or defaults to LEGALENTITYAXIS_DEFAULT if the fact has no LEA
    member.

    :param facts: List of facts to key off by the LegalEntityAxis
    :type facts: list [:class:'arelle.InstanceModelObject.ModelFact']
    :return: Dictionary of a list of facts keyed off of the LegalEntityAxis
    :rtype: dict
    """
    results = defaultdict(list)
    for fact in facts:
        legalDim = LEGALENTITYAXIS_DEFAULT
        if _fact_components_valid(fact):
            dims = [
                dim for dim in fact.context.segDimValues.values()
                if dim.isExplicit and dim.member is not None
            ]
            for dim in dims:
                if dim.dimension.qname.localName == 'LegalEntityAxis':
                    legalDim = dim.member.qname.localName
                    break
            results[legalDim].append(fact)
    return results


def axis_member_fact(axis_name, member_name, model_xbrl):
    """
    Return the fact present, if any, under the axis/member combination.

    :param axis_name: The axis name to check
    :type axis_name: str
    :param member_name: The member name to check
    :type member_name: str
    :return: The fact found or None
    :rtype: :class:'~arelle.InstanceModelObject.ModelFact' or None
    """
    fact_list = []
    for fact in model_xbrl.facts:
        if _fact_components_valid(fact):
            dims = [
                dim for dim in fact.context.segDimValues.values()
                if (
                    dim.isExplicit and
                    dim.member is not None and
                    dim.member.qname.localName == member_name
                )
            ]
            for dim in dims:
                if dim.dimension.qname.localName == axis_name:
                    fact_list.append(fact)
    return fact_list


def _fact_components_valid(fact):
    """
    Return true if all of the components in a fact are not none

    :param fact: The fact to check if it is valid
    :type fact: :class:'~arelle.InstanceModelObject.ModelFact'
    :return: True if none of the components of the fact are not None
    :rtype: bool
    """
    if fact is None:
        return False
    elif fact.context is None:
        return False
    elif fact.context.segDimValues is None:
        return False
    elif fact.context.segDimValues.values() is None:
        return False
    return True


def member_qnames(fact, axis_filter=None):
    """
    Return a list of a fact's member(s)

    :param fact: An arelle ModelFact instance.
    :type fact: :class:'~arelle.InstanceModelObject.ModelFact'
    :param axis_filter: The axis to filter for
    :type axis_filter: bool
    :return: ([str, ..., str]) A list of the string representation of each of
        the fact's member's qnames.
    :rtype: list [str]
    """
    if axis_filter:
        return [
            str(dim.member.qname) for dim in fact.context.segDimValues.values()
            if dim.isExplicit and dim.member is not None and dim.dimensionQname
            is not None and dim.dimensionQname.localName in axis_filter
        ]
    else:
        return [
            str(dim.member.qname) for dim in fact.context.segDimValues.values()
            if dim.isExplicit and dim.member is not None
        ]


def axis_qnames(fact):
    """
    Returns a list the dim.dimensionQnames withing the fact as strings

    :param fact: An arelle ModelFact instance.
    :type fact: :class:'~arelle.InstanceModelObject.ModelFact'
    :return: a list of the fact's axes.
    :rtype: list [str]
    """
    return [
        str(dim.dimensionQname) for dim in fact.context.segDimValues.values()
        if dim.dimensionQname is not None
    ]


def grab_numeric_facts(facts_list):
    """
    Given a list of facts, return those facts whose values are numeric

    :param facts_list: list of fact to return numeric values for
    :type facts_list: list [:class:'~arelle.ModelInstanceObject.ModelFact']
    :return: return list of facts with numeric values
    :rtype: list [:class:'~arelle.ModelInstanceObject.ModelFact']
    """
    numeric_facts = []
    for fact in facts_list:
        if not fact.isNumeric:
            continue
        elif fact.xValue is not None:
            numeric_facts.append(fact)
    return numeric_facts



# (c) Copyright 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
import json
import os

from .util import messages

from arelle import XbrlConst, ModelXbrl
from arelle.ModelDtsObject import ModelConcept
from arelle.FileSource import openFileStream, openFileSource, saveFile


_CODE_NAME = 'DQC.US.0018'
_RULE_VERSION = '2.0.0'
_EARLIEST_US_GAAP_YEAR = 2014

ugtDocs = (
    {
        "year": 2014,
        "namespace": "http://fasb.org/us-gaap/2014-01-31",
        "docLB": "http://xbrl.fasb.org/us-gaap/2014/us-gaap-2014-01-31.zip/us-gaap-2014-01-31/elts/us-gaap-doc-2014-01-31.xml",  # noqa
        "entryXsd": "http://xbrl.fasb.org/us-gaap/2014/us-gaap-2014-01-31.zip/us-gaap-2014-01-31/entire/us-gaap-entryPoint-std-2014-01-31.xsd",  # noqa
    },
    {
        "year": 2015,
        "namespace": "http://fasb.org/us-gaap/2015-01-31",
        "docLB": "http://xbrl.fasb.org/us-gaap/2015/us-gaap-2015-01-31.zip/us-gaap-2015-01-31/elts/us-gaap-doc-2015-01-31.xml",  # noqa
        "entryXsd": "http://xbrl.fasb.org/us-gaap/2015/us-gaap-2015-01-31.zip/us-gaap-2015-01-31/entire/us-gaap-entryPoint-std-2015-01-31.xsd",  # noqa
    },
    {
        "year": 2016,
        "namespace": "http://fasb.org/us-gaap/2016-01-31",
        "docLB": "http://xbrl.fasb.org/us-gaap/2016/us-gaap-2016-01-31.zip/us-gaap-2016-01-31/elts/us-gaap-doc-2016-01-31.xml",  # noqa
        "entryXsd": "http://xbrl.fasb.org/us-gaap/2016/us-gaap-2016-01-31.zip/us-gaap-2016-01-31/entire/us-gaap-entryPoint-std-2016-01-31.xsd",  # noqa
    },
)


def _load_cache(val):
    """
    Loads the needed deprecated concepts cache into memory

    :param val: ValidateXbrl to load the concepts into
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :return: Return True if cache exists, False otherwise.
    :rtype: bool
    """
    cntlr = val.modelXbrl.modelManager.cntlr
    year = _EARLIEST_US_GAAP_YEAR

    for ugt in ugtDocs:
        ugt_namespace = ugt["namespace"]
        if ((ugt_namespace in val.modelXbrl.namespaceDocs and
             len(val.modelXbrl.namespaceDocs[ugt_namespace]) > 0)):
            val.ugtNamespace = ugt_namespace
            deprecations_json_file = os.path.join(
                os.path.dirname(__file__),
                'resources',
                'DQC_US_0018',
                '{}_deprecated-concepts.json'.format(str(year))
            )
            file = None

            try:
                file = openFileStream(
                    cntlr,
                    deprecations_json_file,
                    'rt',
                    encoding='utf-8'
                )

                val.usgaapDeprecations = json.load(file)
                file.close()

            except FileNotFoundError:  # noqa
                if file:
                    file.close()
                # year should be cached. It is not, so return False
                return False
            # year should be cached, and is. Return True
            return True
        year += 1
    # checked all years. No cache found.
    return False


def _create_cache(val):
    """
    Creates the caches needed for dqc_us_0018

    :param val: ValidateXbrl needed in order to save the cache
    :type val: :class: '~arelle.ValidateXbrl.ValidateXbrl'
    :return: no explicit return but creates and saves a cache in
        dqc_us_rule\resources\DQC_US_0018
    :rtype: None
    """
    val.ugtNamespace = None
    cntlr = val.modelXbrl.modelManager.cntlr
    year = _EARLIEST_US_GAAP_YEAR

    for ugt in ugtDocs:
        deprecations_json_file = os.path.join(
            os.path.dirname(__file__),
            'resources',
            'DQC_US_0018',
            '{}_deprecated-concepts.json'.format(str(year))
        )

        if not os.path.isfile(deprecations_json_file):
            ugt_doc_lb = ugt["docLB"]
            val.usgaapDeprecations = {}
            disclosure_system = (
                val.modelXbrl.modelManager.validateDisclosureSystem
            )

            prior_validate_disclosure_system = disclosure_system
            val.modelXbrl.modelManager.validateDisclosureSystem = False
            deprecations_instance = ModelXbrl.load(
                val.modelXbrl.modelManager,
                openFileSource(ugt_doc_lb, cntlr),
                _("built deprecations table in cache")  # noqa
            )
            val.modelXbrl.modelManager.validateDisclosureSystem = (
                prior_validate_disclosure_system
            )

            if deprecations_instance is not None:
                dep_label = 'http://www.xbrl.org/2009/role/deprecatedLabel'
                dep_date_label = (
                    'http://www.xbrl.org/2009/role/deprecatedDateLabel'
                )
                concept_label = XbrlConst.conceptLabel
                relationship_set = (
                    deprecations_instance.relationshipSet(concept_label)
                )
                model_relationships = relationship_set.modelRelationships

                for labelRel in model_relationships:
                    model_documentation = labelRel.toModelObject
                    concept = labelRel.fromModelObject.name

                    if model_documentation.role == dep_label:
                        val.usgaapDeprecations[concept] = (
                            model_documentation.text,
                            val.usgaapDeprecations.get(concept, ('', ''))[0]
                        )
                    elif model_documentation.role == dep_date_label:
                        val.usgaapDeprecations[concept] = (
                            model_documentation.text,
                            val.usgaapDeprecations.get(concept, ('', ''))[1]
                        )
                json_str = str(
                    json.dumps(
                        val.usgaapDeprecations,
                        ensure_ascii=False, indent=0
                    )
                )  # might not be unicode in 2.7
                saveFile(cntlr, deprecations_json_file, json_str)
                deprecations_instance.close()
                del deprecations_instance  # dereference closed modelXbrl
        year += 1


def deprecated_concept_errors(val, *args, **kwargs):
    """
    Makes error messages for all deprecation errors

    :param val: ValidateXbrl to check for error
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :return: No explicit return, though error messages are created
    :rtype: None
    """
    cache_exists = _load_cache(val)
    if not cache_exists:
        _create_cache(val)

    deprecated_concepts = {}
    _catch_deprecated_fact_errors(val, deprecated_concepts)
    _catch_linkbase_deprecated_errors(val, deprecated_concepts)

    for key in deprecated_concepts.keys():
        val.modelXbrl.error(
            '{}.34'.format(_CODE_NAME),
            messages.get_message(_CODE_NAME),
            concept=key,
            modelObject=deprecated_concepts[key],
            ruleVersion=_RULE_VERSION
        )


def _deprecated_concept(val, concept):
    """
    Returns true if the fact uses a deprecated concept

    :param val: ValidateXbrl to check for deprecated items
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :param concept: Concept to check
    :type concept: :class:'~arelle.ModelFact.Concept'
    :return: Returns true if the fact uses a deprecated concept
    :rtype: bool
    """
    if not isinstance(concept, ModelConcept):
        return False
    elif ((hasattr(val, 'usgaapDeprecations') and
           concept.name in val.usgaapDeprecations)):
        return True
    return False


def _deprecated_dimension(val, dim_concept):
    """
    Returns true if the fact uses a deprecated dimension

    :param val: ValidateXbrl to check for deprecated dimensions
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :param dim_concept: Concept to check
    :type dim_concept: :class:'~arelle.ModelDtsObject.ModelConcept'
    :return: Returns true if the fact uses a deprecated dimension
    :rtype: bool
    """
    if not isinstance(dim_concept, ModelConcept):
        return False
    elif ((hasattr(val, 'usgaapDeprecations') and
           dim_concept.name in val.usgaapDeprecations)):
        return True
    return False


def _deprecated_member(val, model_dim):
    """
    Returns true if the fact uses a deprecated member

    :param val: ValidateXbrl to check for deprecated dimensions
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :param modelDim: Concept to check
    :type modelDim: :class:'~arelle.ModelInstanceObject.ModelDimensionValue'
    :return: Returns true if the fact uses a deprecated dimension
    :rtype: bool
    """
    if model_dim.isExplicit:
        member = model_dim.member
        if member is not None and hasattr(val, 'usgaapDeprecations'):
            if member.name in val.usgaapDeprecations:
                return True
    return False


def _fact_checkable(fact):
    """
    Returns true if fact can be checked

    :param fact: Fact to check concept and context for None
    :type fact: :class:'~arelle.ModelDtsObject.ModelFact'
    :return: Returns true if the fact can be checked
    :rtype: bool
    """
    return (
        fact.concept is not None and
        fact.context is not None
    )


def _catch_linkbase_deprecated_errors(val, deprecated_concepts):
    """
    Check for unused concept relationships of standard taxonomy elements
    and catches abstract deprecated concepts in linkbases

    :param val: :class: '~arelle.ValdiateXbrl.ValidateXbrl'
    :return: No Return
    :rtype: None
    """
    relationships = val.modelXbrl.relationshipSet(XbrlConst.parentChild)
    for rel in relationships.modelRelationships:
        for concept in (rel.fromModelObject, rel.toModelObject):
            if _deprecated_concept(val, concept):
                if not deprecated_concepts.get(concept.name):
                    deprecated_concepts[concept.name] = []
                deprecated_concepts[concept.name].append(
                    rel.locatorOf(concept)
                )


def _fact_uses_deprecated_item(val, fact):
    """
    Checks to see if a fact uses a deprecated item

    :param val: ValidateXbrl to check for deprecated items
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :param fact: Fact to check
    :type fact: :class:'~arelle.ModelInstanceObject.ModelFact'
    :return: Returns true if fact uses deprecated item,
        as well as the item's name.
    :rtype: tuple(bool, str)
    """

    if _fact_checkable(fact):
        if _deprecated_concept(val, fact.concept):
            return True, fact.concept.name

        if fact.isItem:
            for dimConcept, modelDim in fact.context.segDimValues.items():
                if _deprecated_concept(val, dimConcept):
                    return True, dimConcept.name
                elif _deprecated_dimension(val, modelDim.dimension):
                    return True, modelDim.dimension.name
    return False, None


def _catch_deprecated_fact_errors(val, deprecated_concepts):
    """
    Checks to see if facts are using deprecated items
    :param val: ValidateXbrl to check for deprecated item

    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :return: No Return
    :rype: None
    """
    for fact in val.modelXbrl.facts:
        fire, item = _fact_uses_deprecated_item(val, fact)
        if fire:
            if not deprecated_concepts.get(item):
                deprecated_concepts[item] = []
            deprecated_concepts[item].append(fact)


__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': (
        'Checks all of the specified types and concepts for their '
        'date ranges to verify the ranges are within expected '
        'parameters for the fiscal periods'
    ),
    # Mount points
    'Validate.XBRL.Finally': deprecated_concept_errors
}



# (c) Copyright 2015 - 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
import pkgutil
import sys
import inspect


def run_checks(val, *args, **kwargs):
    """
    The function to run all of the validation checks under the SEC package
    """
    plugin_modules = _plugins_to_run(sys.modules[__name__])
    for plugin in plugin_modules:
        try:
            if ((plugin.__file__ is not None and
                 plugin.__file__.find('__init__.py') == -1 and
                 hasattr(plugin, '__pluginInfo__'))):

                func = plugin.__pluginInfo__['Validate.XBRL.Finally']
                func(val, *args, **kwargs)
        except Exception as err:
            # This is an overly generic error catch, but it will hopefully
            # be able to be pared down in the future.
            val.modelXbrl.error(
                "dqc_us_rules.exception:" + type(err).__name__,
                (
                    "Testcase validation exception: "
                    "%(error)s, testcase: %(testcase)s"
                ),
                modelXbrl=val.modelXbrl,
                testcase=val.modelXbrl.modelDocument.basename,
                error=err,
                exc_info=True
            )


def _plugins_to_run(mod, include_start=True):
    """
    Accepts a module/package and returns an iterator
    that will yield all the submodules (or subpackages)
    of the module/package.

    Set `include_start=False` if you do not want the
    iterator to yield the initial, passed module.
    """
    if inspect.ismodule(mod):
        if include_start:
            yield(mod)

        path = getattr(mod, '__path__', None)
        prefix = mod.__name__ + "."
        if path is not None:
            for importer, modname, _ in pkgutil.iter_modules(path):
                sub_mod = __import__(prefix + modname, fromlist="dummy")
                for m in _plugins_to_run(sub_mod):
                    yield m


__pluginInfo__ = {
    'name': 'DQC.SEC.ALL',
    'version': '2.0',
    'description': 'All Data Quality Committee SEC Filing Checks',
    'author': '',
    'license': 'See accompanying license text',
    # Required plugin for logging
    'import': ('logging/dqcParameters.py', ),
    # Mount points
    'Validate.XBRL.Finally': run_checks,
}



# (c) Copyright 2015 - 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
import re
from .util import facts, messages

from arelle.XmlUtil import dateunionValue

LEGALENTITYAXIS_DEFAULT = ''
_dei_pattern = (
    re.compile(r'^http://xbrl\.((sec\.gov)|(us))/dei/\d{4}-\d{2}-\d{2}')
)
_CODE_NAME = 'DQC.US.0005'
_RULE_VERSION = '2.1.0'

_REPORT_TYPE_EXCLUSIONS = ['S-1', 'S-3', 'S-4', 'S-6', 'S-8', 'S-11', 'S-20', 'S-1/A', 'S-3/A', 'S-4/A', 'S-6/A', 'S-8/A', 'S-11/A', 'S-20/A']  # noqa


def _get_end_of_period(val):
    """
    Gets the end of period date (dei:DocumentPeriodEndDate).  Returns the fact
    value or the context end date on the fact, whichever is later, in a tuple
    with the fact itself and a string representation of the date. The tuple is
    in a dict keyed off of any LegalEntityAxis members. The end of period
    string is adjusted for viewing to represent the day as expected.  In other
    words, dates that end in 24:00 will be put at 00:00 of the expected day.

    :param val: bal from which to gather end dates
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: A dictionary of tuples containing the fact, found date and a
        string representation of that date, keyed off of the LegalEntityAxis
        members in the format {lea_member: (fact, found_date, date_str)}
    :rtype: dict
    """
    results = {}
    end_of_period_concepts = [
        c for c in val.modelXbrl.nameConcepts['DocumentPeriodEndDate']
        if c.qname.namespaceURI in val.disclosureSystem.standardTaxonomiesDict
    ]

    if len(end_of_period_concepts) == 1:
        end_of_period_dict = (
            facts.legal_entity_axis_facts_by_member(
                val.modelXbrl.factsByQname[end_of_period_concepts[0].qname]
            )
        )
        for lea_member, end_of_period_facts in end_of_period_dict.items():
            for fact in end_of_period_facts:
                eop_date = fact.xValue
                # Get maximum of fact value and fact's context end date
                if fact.context is not None:
                    eop_context_end = fact.context.endDatetime
                    date_str = (
                        dateunionValue(eop_context_end, subtractOneDay=True)
                    )
                    if ((eop_context_end is not None and
                         (eop_date is None or eop_context_end > eop_date))):
                        eop_date = eop_context_end
                        # end dates have a time of 24:00 so
                        # adjust them back 1 day
                        date_str = (
                            dateunionValue(eop_date, subtractOneDay=True)
                        )
                    if lea_member in results:
                        if results[lea_member][1] < eop_date:
                            results[lea_member] = (fact, eop_date, date_str)
                    else:
                        results[lea_member] = (fact, eop_date, date_str)
    return results


def validate_facts(val, *args, **kwargs):
    """
    This function validates facts. In other words this function checks to see
    if the facts contained in val are correctly implemented. Ignores these
    checks with any forms in the _REPORT_TYPE_EXCLUSIONS list.

    :param val: val to check
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: No direct return, throws errors when facts can't be validated
    :rtype: None
    """
    # Ignore checks if this document is in the REPORT_TYPE_EXCLUSIONS list.
    dei_list = facts.get_facts_dei(['DocumentType'], model_xbrl=val.modelXbrl)
    for dei in dei_list:
        is_an_excluded_report = any(
            True if dei.xValue is not None and ex in str(dei.xValue) else False
            for ex in _REPORT_TYPE_EXCLUSIONS
        )
        if is_an_excluded_report:
            return

    eop_results = _get_end_of_period(val)
    fact_dict = facts.legal_entity_axis_facts_by_member(
        filter(lambda f: f.context is not None, val.modelXbrl.facts)
    )
    for lea_member, fact_list in fact_dict.items():
        lookup = (
            lea_member
            if lea_member in eop_results
            else facts.LEGALENTITYAXIS_DEFAULT
        )
        if lookup in sorted(eop_results):
            for fact in fact_list:
                # endDateTime will be the instant date time if this
                # is an instant period
                fact_date = fact.context.endDatetime
                if fact_date is not None:
                    # Check for the case where a fact is less than
                    # the expected eop dates
                    comparison_date = eop_results[lookup][1]
                    if fact_date <= comparison_date:
                        run_checks(val, fact, eop_results, lookup)


def run_checks(val, fact, eop_results, lookup):
    """
    Called to determine which error to fire.

    :param val: val to check
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param fact: fact to check
    :type fact: :class:'~arelle.ModelInstanceObject.ModelFact'
    :param eop_results: A dictionary of tuples containing the fact, found
        date and a string representation of that date, keyed off of the
        LegalEntityAxis members in the format:
            {lea_member: (fact, found_date, date_str)}
    :type eop_results: dict
    :param lookup: legal entity access member concept.
    :type lookup: str
    :return: No direct return, throws errors when facts can't be validated
    :rtype: None
    """
    if fact.qname.localName == 'EntityCommonStockSharesOutstanding':
        fact_date = fact.context.endDatetime
        comparison_date = eop_results[lookup][1]
        # if a fact whose qname is
        # EntityCommonStockSharesOutstanding has an
        # end date prior to eop then fire an error

        # Only fire if it is actually less than the
        # comparison date
        if fact_date < comparison_date:
            val.modelXbrl.error(
                '{base_code}.17'.format(
                    base_code=_CODE_NAME
                ),
                messages.get_message(_CODE_NAME, "17"),
                modelObject=[fact] +
                list(eop_results[lookup]),
                ruleVersion=_RULE_VERSION
            )
    elif facts.axis_exists(
            val, fact, 'SubsequentEventTypeAxis'
    ):
        val.modelXbrl.error(
            '{base_code}.48'.format(base_code=_CODE_NAME),
            messages.get_message(_CODE_NAME, "48"),
            modelObject=[fact] + list(eop_results[lookup]),
            ruleVersion=_RULE_VERSION
        )
    elif facts.axis_member_exists(
            val,
            fact,
            'StatementScenarioAxis',
            'ScenarioForecastMember'
    ):
        val.modelXbrl.error(
            '{base_code}.49'.format(base_code=_CODE_NAME),
            messages.get_message(_CODE_NAME, "49"),
            modelObject=[fact] + list(eop_results[lookup]),
            ruleVersion=_RULE_VERSION
        )


__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': 'A subsequent event test ',
    # Mount points
    'Validate.XBRL.Finally': validate_facts,
}



# (c) Copyright 2015 - 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
from decimal import Decimal
from math import isnan
from .util import facts, messages
from arelle.ValidateXbrlCalcs import inferredDecimals, roundValue

_ASSETS_CONCEPT = 'Assets'
_LIABILITIES_CONCEPT = 'LiabilitiesAndStockholdersEquity'
_CODE_NAME = 'DQC.US.0004'
_RULE_VERSION = '1.0.0'


def assets_eq_liability_equity(val, *args, **kwargs):
    """
    Assets equals Liabilities and Stockholders Equity

    :param val: val to look at the modelXbrl for
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: No direct return, but throws an error for everything returned by
        _assets_eq_liability_equity
    :rtype: None
    """
    for fact in _assets_eq_liability_equity(val.modelXbrl):
        fact_assets, fact_liabilities = fact
        val.modelXbrl.error(
            '{}.16'.format(_CODE_NAME),
            messages.get_message(_CODE_NAME),
            modelObject=[fact_assets, fact_liabilities],
            ruleVersion=_RULE_VERSION
        )


def _assets_eq_liability_equity(model_xbrl):
    """
    Yields fact assets and fact liabilities as long as it is able to

    :param model_xbrl: modelXbrl to check name concepts of
    :type model_xbrl: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: yields fact assets and fact liabilities that should throw errors
    :rtype: tuple
    """
    assets_concept = (
        model_xbrl.nameConcepts[_ASSETS_CONCEPT][0]
        if model_xbrl.nameConcepts[_ASSETS_CONCEPT] else None
    )

    liability_equity_concept = (
        model_xbrl.nameConcepts[_LIABILITIES_CONCEPT][0]
        if model_xbrl.nameConcepts[_LIABILITIES_CONCEPT]
        else None
    )

    if assets_concept is not None and liability_equity_concept is not None:
        assets_facts = model_xbrl.factsByQname[assets_concept.qname]
        liability_equity_facts = (
            model_xbrl.factsByQname[liability_equity_concept.qname]
        )

        fact_dict = dict()
        fact_dict[_ASSETS_CONCEPT] = assets_facts
        fact_dict[_LIABILITIES_CONCEPT] = liability_equity_facts
        fact_groups = facts.prepare_facts_for_calculation(fact_dict)

        for fact_group in fact_groups:
            fact_assets = fact_group[_ASSETS_CONCEPT]
            fact_liabilities = fact_group[_LIABILITIES_CONCEPT]
            if ((fact_assets.context is not None and
                 fact_assets.context.instantDatetime is not None)):
                dec_assets = inferredDecimals(fact_assets)
                dec_liabilities = inferredDecimals(fact_liabilities)
                min_dec = min(dec_assets, dec_liabilities)
                if _values_unequal(
                    fact_assets.xValue, fact_liabilities.xValue, min_dec
                ):
                    yield fact_assets, fact_liabilities


def _min_dec_valid(min_dec):
    """
    Checks to make sure that min_dec values are valid.
    Returns False if min_dec is a None type or if min_dec is Not a number

    :param min_dec: the minumum of dec_assets and dec_liabilities
    :type min_dec: None or int or NaN
    :return: Is min_dec valid
    :rtype: bool
    """

    return min_dec is not None and not isnan(min_dec)


def _values_unequal(val1, val2, dec_scale, margin_scale=2):
    """
    Checks the values for equality based on their scaling.
    Returns False if the values are equal, otherwise True.

    :param val1: first value to round
    :type val1: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param val2: second value to round
    :type val2: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param dec_scale: precision on rounded value
    :type dec_scale: :class:'~decimal.Decimal'
    :param margin_scale: margin of scale for the margin of error
    :type margin_scale: float
    :return: True if the values are not equal
    :rtype: bool
    """

    if not _min_dec_valid(dec_scale):
        return False

    round_val1 = roundValue(val1, decimals=dec_scale)
    round_val2 = roundValue(val2, decimals=dec_scale)
    margin_of_error = (
        Decimal(margin_scale) * (Decimal(10) ** Decimal(-dec_scale))
    )
    return (
        round_val1 < round_val2 - margin_of_error or
        round_val1 > round_val2 + margin_of_error
    )


__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': 'Assets should equal Liabilities and Shareholders Equity',
    # Mount points
    'Validate.XBRL.Finally': assets_eq_liability_equity,
}



# (c) Copyright 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
from .util import messages

import json
import os
import re
import time

from arelle.ModelDtsObject import ModelConcept
from arelle import XbrlConst, ModelXbrl
from arelle.FileSource import openFileStream, saveFile, openFileSource


_CODE_NAME = 'DQC.US.0041'
_RULE_VERSION = '2.0.0'

_EARLIEST_GAAP_YEAR = 2014

ugtDocs = (
    {
        "year": 2014,
        "namespace": "http://fasb.org/us-gaap/2014-01-31",
        "docLB": "http://xbrl.fasb.org/us-gaap/2014/elts/us-gaap-doc-2014-01-31.xml",  # noqa
        "entryXsd": "http://xbrl.fasb.org/us-gaap/2014/entire/us-gaap-entryPoint-std-2014-01-31.xsd",  # noqa
     },

    {
        "year": 2015,
        "namespace": "http://fasb.org/us-gaap/2015-01-31",
        "docLB": "http://xbrl.fasb.org/us-gaap/2015/us-gaap-2015-01-31.zip/us-gaap-2015-01-31/elts/us-gaap-doc-2015-01-31.xml",  # noqa
        "entryXsd": "http://xbrl.fasb.org/us-gaap/2015/us-gaap-2015-01-31.zip/us-gaap-2015-01-31/entire/us-gaap-entryPoint-std-2015-01-31.xsd",  # noqa
     },

    {
        "year": 2016,
        "namespace": "http://fasb.org/us-gaap/2016-01-31",
        "docLB": "http://xbrl.fasb.org/us-gaap/2016/us-gaap-2016-01-31.zip/us-gaap-2016-01-31/elts/us-gaap-doc-2016-01-31.xml",  # noqa
        "entryXsd": "http://xbrl.fasb.org/us-gaap/2016/us-gaap-2016-01-31.zip/us-gaap-2016-01-31/entire/us-gaap-entryPoint-std-2016-01-31.xsd",  # noqa
     }
)


def _make_cache(val, ugt, cntlr, ugt_default_dimensions_json_file):
    """
    Creates a new caches for the Taxonomy default dimensions

    :param val: ValidateXbrl to be validated
    :type val: :class: '~arelle.ValidateXbrl.ValidateXbrl'
    :param ugt: Taxonomy to check
    :type ugt: str
    :param ugt_default_dimensions_json_file: location to save json default
        dimensions
    :type ugt_default_dimensions_json_file: str
    :return: no explicit return, but saves caches for dqc_us_0041
    :rtype: None
    """
    started_at = time.time()
    ugt_entry_xsd = ugt["entryXsd"]
    val.usgaapDefaultDimensions = {}
    prior_validate_disclosure_system = (
        val.modelXbrl.modelManager.validateDisclosureSystem
    )
    val.modelXbrl.modelManager.validateDisclosureSystem = False
    ugt_entry_xsd_instance = (
        ModelXbrl.load(
            val.modelXbrl.modelManager,
            openFileSource(ugt_entry_xsd, cntlr),
            _("opened us-gaap entry xsd")  # noqa
        )
    )
    val.modelXbrl.modelManager.validateDisclosureSystem = (
        prior_validate_disclosure_system
    )

    if ugt_entry_xsd_instance is None:
        val.modelXbrl.error(
            "arelle:notLoaded",
            _("US-GAAP entry xsd not loaded: %(file)s"),  # noqa
            modelXbrl=val,
            file=os.path.basename(ugt_entry_xsd)
        )

    else:
        model_relationships = (
            ugt_entry_xsd_instance.relationshipSet(
                XbrlConst.dimensionDefault
            ).modelRelationships
        )
        for default_dim_rel in model_relationships:
            if _default_dim_rel_is_instance(default_dim_rel):
                from_name = default_dim_rel.fromModelObject.name
                to_name = default_dim_rel.toModelObject.name
                val.usgaapDefaultDimensions[from_name] = to_name
        json_str = str(
            json.dumps(
                val.usgaapDefaultDimensions,
                ensure_ascii=False,
                indent=0
            )
        )  # might not be unicode in 2.7
        # 2.7 gets unicode this way
        saveFile(cntlr, ugt_default_dimensions_json_file, json_str)
        ugt_entry_xsd_instance.close()
        del ugt_entry_xsd_instance  # dereference closed modelXbrl
    val.modelXbrl.profileStat(
        _("build default dimensions cache"),  # noqa
        time.time() - started_at
    )


def _setup_cache(val):
    """
    Builds the cache needed for dqc_us_0041. Should only have to build it the
    first time

    :param val: ValidateXbrl to check if it contains errors
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :return: No explicit return, but loads the default dimensions of taxonomies
        into memory
    :rtype: None
    """
    val.linroleDefinitionIsDisclosure = (
        re.compile(r"-\s+Disclosure\s+-\s", re.IGNORECASE)
    )

    val.linkroleDefinitionStatementSheet = (
        re.compile(r"[^-]+-\s+Statement\s+-\s+.*", re.IGNORECASE)
    )  # no restriction to type of statement

    val.ugtNamespace = None
    cntlr = val.modelXbrl.modelManager.cntlr

    year = _EARLIEST_GAAP_YEAR
    for ugt in ugtDocs:
        ugt_default_dimensions_json_file = os.path.join(
            os.path.dirname(__file__),
            'resources',
            'DQC_US_0041',
            '{}_ugt-default-dimensions.json'.format(str(year))
        )

        _make_cache(val, ugt, cntlr, ugt_default_dimensions_json_file)

        year += 1


def _is_in_namespace(val, ugt_namespace):
    """
    Returns true if the ugt_namespace is the same as the current vals namespace

    :param val: ValidateXbrl that is being validated
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl
    :param ugt_namespace: Namespace of taxonomy for current year
    :type ugt_namespace: str
    :return: True if ugt_namespace is in the val.modelXbrl's namespace docs
    :rtype: bool
    """
    return (
        ugt_namespace in val.modelXbrl.namespaceDocs and
        len(val.modelXbrl.namespaceDocs[ugt_namespace]) > 0
    )


def _default_dim_rel_is_instance(default_dim_rel):
    """
    Checks to makes sure that the fromModelObject and the toModelObject's of a
    defaultDimensionRelationship are both ModelConcept's

    :param default_dim_rel: The default dimension relationship to check
    :type default_dim_rel: :class:'~arelle.ModelXbrl.ModelRelationship'
    :return: True if the default_dim_rel.fromModelObject is a ModelConcept and
        the default_dim_rel.toModelObject is a ModelConcept
    :rtype: bool
    """
    return(
        isinstance(default_dim_rel.fromModelObject, ModelConcept) and
        isinstance(default_dim_rel.toModelObject, ModelConcept)
    )


def _load_cache(val):
    """
    Loads the cached taxonomy default demensions. If the file isn't cached yet
    it will create a new cache

    :param val: ValidateXbrl to be validated
    :type val: :class: '~arelle.ValidateXbrl.ValidateXbrl'
    :return: no explicit return, but loads caches for dqc_us_0041
    :rtype: None
    """
    val.linroleDefinitionIsDisclosure = (
        re.compile(r"-\s+Disclosure\s+-\s", re.IGNORECASE)
    )

    val.linkroleDefinitionStatementSheet = (
        re.compile(r"[^-]+-\s+Statement\s+-\s+.*", re.IGNORECASE)
    )  # no restriction to type of statement

    val.ugtNamespace = None
    cntlr = val.modelXbrl.modelManager.cntlr

    year = _EARLIEST_GAAP_YEAR

    for ugt in ugtDocs:
        ugt_namespace = ugt["namespace"]
        if _is_in_namespace(val, ugt_namespace):
            ugt_default_dimensions_json_file = os.path.join(
                os.path.dirname(__file__),
                'resources',
                'DQC_US_0041',
                '{}_ugt-default-dimensions.json'.format(str(year))
            )

            file = None

            try:
                file = openFileStream(
                    cntlr, ugt_default_dimensions_json_file,
                    'rt', encoding='utf-8'
                )

                val.usgaapDefaultDimensions = json.load(file)
                file.close()

            except FileNotFoundError:  # noqa
                if file:
                    file.close()
        year += 1


def fire_dqc_us_0041_errors(val, *args, **kwargs):
    """
    Fires all the dqc_us_0041 errors returned by _catch_dqc_us_0041_errors

    :param val: ValidateXbrl to check if it contains errors
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :return: No explicit return, but it fires all the dqc_us_0041 errors
    :rtype: None
    """
    _load_cache(val)

    for error_info in _catch_dqc_us_0041_errors(val):
        axis_name, axis_default_name, def_name = error_info
        val.modelXbrl.error(
            '{}.73'.format(_CODE_NAME),
            messages.get_message(_CODE_NAME),
            axis_name=axis_name,
            axis_default_name=axis_default_name,
            def_name=def_name,
            ruleVersion=_RULE_VERSION
        )


def _check_relationship_exists(rel):
    """
    Makes sure that the relationship won't throw None type errors

    :param rel: relationship to check
    :type rel: :class:'~arelle.ModelXbrl.ModelRelationship'
    :return: Returns true if both the toModelObject and the fromModelObject
        are both instances of ModelConcepts
    :rtype: bool
    """
    return (
        isinstance(rel.toModelObject, ModelConcept) and
        isinstance(rel.fromModelObject, ModelConcept)
    )


def _default_dimension_mismatch(relation, validation):
    """
    Returns true if the default dimension is not included in the usgaap default
    dimensions

    :param relation: dimension to test if it is a usgaap default dimension.
    :type relation: :class:`~arelle.ModelDtsObject.ModelRelationship`
    :param validation: The validation object used to look up the default name.
    :type validation: :class:`~arelle.ValidationObject.ValidateXbrl`
    :return: True if the default dimensions is not included in the usgaap
        default dimensions.
    :rtype: bool
    """
    if ((not hasattr(validation, "usgaapDefaultDimensions") or
         not isinstance(validation.usgaapDefaultDimensions, dict))):
        return False
    has_a_default_name = (
        validation.usgaapDefaultDimensions.get(relation.fromModelObject.name)
        is not None
    )
    is_the_default_name = (
        relation.toModelObject.name !=
        validation.usgaapDefaultDimensions.get(relation.fromModelObject.name)
    )
    if has_a_default_name and is_the_default_name:
        return True
    return False


def _catch_dqc_us_0041_errors(val):
    """
    Returns a tuple containing the parts of the dqc_us_0041 error to be
    displayed

    :param val: ValidateXbrl to check if it contains errors
    :type val: :class:'~arelle.ValidateXbrl.ValidateXbrl'
    :return: all dqc_us_0041 errors
    :rtype: tuple
    """
    _load_cache(val)
    rel_set = val.modelXbrl.relationshipSet(
        XbrlConst.dimensionDefault
    ).modelRelationships
    for relation in rel_set:
        if _check_relationship_exists(relation):
            if _default_dimension_mismatch(relation, val):
                yield (
                    relation.fromModelObject.name,
                    val.usgaapDefaultDimensions.get(
                        relation.fromModelObject.name
                    ),
                    relation.toModelObject.name
                )

__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': (
        'All axis defaults should be the same as the axis '
        'defaults defined in the taxonomy.'
    ),
    # Mount points
    'Validate.XBRL.Finally': fire_dqc_us_0041_errors,
}



# (c) Copyright 2015 - 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
import os
import collections

from .util import facts, messages, neg_num


_CODE_NAME = 'DQC.US.0013'
_RULE_VERSION = '2.0.0'
_DEFAULT_CONCEPTS_FILE = os.path.join(
    os.path.dirname(__file__),
    'resources',
    'DQC_US_0013',
    'dqc_13_concepts.csv'
)
# The same exclusion rules used in Rule 15 also apply to this rule
_DEFAULT_EXCLUSIONS_FILE = os.path.join(
    os.path.dirname(__file__),
    'resources',
    'DQC_US_0015',
    'dqc_15_exclusion_rules.csv'
)

# This is setup as a list of tuples.  The first item is the concept to check
# for existence of, the second is a list of concept(s) to sum and check if
# they are greater than zero.
_PRE_CFG = [
    (
        'IncomeLossFromContinuingOperationsBeforeIncomeTaxesExtraordinaryItemsNoncontrollingInterest',  # noqa
        [
            'IncomeLossFromContinuingOperationsBeforeIncomeTaxesExtraordinaryItemsNoncontrollingInterest'  # noqa
        ]
    ),
    (
        'IncomeLossFromContinuingOperationsBeforeIncomeTaxesMinorityInterestAndIncomeLossFromEquityMethodInvestments',  # noqa
        [
            'IncomeLossFromContinuingOperationsBeforeIncomeTaxesMinorityInterestAndIncomeLossFromEquityMethodInvestments',  # noqa
            'IncomeLossFromEquityMethodInvestments'
        ]
    ),
    (
        'IncomeLossFromContinuingOperationsBeforeIncomeTaxesDomestic',
        [
            'IncomeLossFromContinuingOperationsBeforeIncomeTaxesDomestic',
            'IncomeLossFromContinuingOperationsBeforeIncomeTaxesForeign',
            'IncomeLossFromEquityMethodInvestments'
        ]
    ),
    (
        'IncomeLossFromContinuingOperationsBeforeIncomeTaxesForeign',
        [
            'IncomeLossFromContinuingOperationsBeforeIncomeTaxesDomestic',
            'IncomeLossFromContinuingOperationsBeforeIncomeTaxesForeign',
            'IncomeLossFromEquityMethodInvestments'
        ]
    )
]

_PRECONDITION_ELEMENTS = collections.OrderedDict(_PRE_CFG)


def run_negative_values_with_dependence(val, *args, **kwargs):
    """
    Run the list of facts against our negative number checks and add errors for
    the hits in the various lists.

    :param val: The validation object which carries the validation information,
        including the ModelXBRL
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: No direct return, but throws errors for facts matching the
        blacklist
    :rtype: None
    """
    # filter down to numeric facts
    blacklist_dict = neg_num.concept_map_from_csv(_DEFAULT_CONCEPTS_FILE)
    blacklist_facts = filter_negative_number_with_dependence_facts(
        val, blacklist_dict.keys()
    )
    for fact in blacklist_facts:
        index_key = blacklist_dict[fact.qname.localName]
        val.modelXbrl.error(
            '{base_key}.{extension_key}'.format(
                base_key=_CODE_NAME, extension_key=index_key
            ),
            messages.get_message(_CODE_NAME, str(index_key)),
            concept=fact.concept.label(), modelObject=fact,
            ruleVersion=_RULE_VERSION
        )


def filter_negative_number_with_dependence_facts(val, blacklist_concepts):
    """
    Checks the numeric, negative value facts in the provided ModelXBRL instance
    against the rule dictionary and returns those which meet the conditions of
    the blacklist.

    :param val: val whose modelXbrl provides the facts to check
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param blacklist_concepts: An iterable of the blacklist concepts we should
        be testing against.
    :type blacklist_concepts: list [str]
    :return: Return list of the facts falling into the blacklist.
    :rtype: list [:class:'~arelle.ModelInstanceObject.ModelFact']
    """
    blacklist_exclusion_rules = neg_num.get_rules_from_csv(
        _DEFAULT_EXCLUSIONS_FILE
    )
    bad_blacklist = []
    numeric_facts = facts.grab_numeric_facts(list(val.modelXbrl.facts))
    # Checks if the precondition concept exists and only proceeds with check
    # if true and if the fact context matches the context of the
    # precondition fact
    precondition_contexts = dqc_13_precondition_check(val)

    # other filters before running negative numbers check
    # numeric_facts has already checked if fact.value can be made into
    # a number
    if len(precondition_contexts):
        facts_to_check = [
            fact for fact in numeric_facts
            if fact.xValue < 0 and
            fact.context in precondition_contexts
        ]
    else:
        facts_to_check = []
    # identify facts which should be reported as included in the list
    for fact in facts_to_check:
        if neg_num.check_rules(fact, blacklist_exclusion_rules):
            continue  # cannot be black
        if fact.qname.localName in blacklist_concepts:
            bad_blacklist.append(fact)
    return bad_blacklist


def dqc_13_precondition_check(val):
    """
    Checks if the precondition fact(s) exist and grabs their values.  Runs
    through an if/else statement of the precondition check and totals the
    values to ensure they are greater than zero.

    :param val: val whose modelXbrl provides the facts to check
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: Returns the list of context(s) of the precondition fact(s)
    :rtype: list [:class:'~arelle.ModelInstanceObject.ModelContext']

    """
    check_contexts = set()
    seen_contexts = set()
    pre_cache = {
        x: {
            y: [] for y in pre_checks
        } for x, pre_checks in _PRECONDITION_ELEMENTS.items()
    }
    for fact in list(val.modelXbrl.facts):
        for precondition, pre_check_dict in pre_cache.items():
            for pre_check in pre_check_dict:
                if fact.concept.qname.localName == pre_check:
                    pre_check_dict[pre_check].append(fact)
    for precondition_name in _PRECONDITION_ELEMENTS.keys():
        precondition_contexts = set()  # To keep track of preconditions
        context_group_dict = collections.defaultdict(list)
        for concept_name, fact_list in pre_cache[precondition_name].items():
            for fact in fact_list:
                context_group_dict[fact.context].append(fact)
                if concept_name == precondition_name:
                    precondition_contexts.add(fact.context)
        for context, fact_list in context_group_dict.items():
            if context in seen_contexts:
                # We have already found a rule for this context, ignored
                continue
            if context not in precondition_contexts:
                # This context does not have the precondition element, ignored
                continue
            value = sum(filter(None, [f.xValue for f in fact_list]))
            if value > 0:
                check_contexts.add(context)
            else:
                # We've already checked this context so no need to check again
                seen_contexts.add(context)
    return check_contexts


__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': (),
    # Mount points
    'Validate.XBRL.Finally': run_negative_values_with_dependence,
}



import csv
from . import facts


def check_rules(fact, rule_dicts):
    for rule_dict in rule_dicts:
        if check_rule(fact, rule_dict):
            return True
    return False


def check_rule(fact, rule_dict):
    """
    Check if the input fact meets the conditions of the passed in rule_dict.

    :param fact: An arelle ModelFact instance.
    :type fact: :class:'~arelle.ModelInstanceObject.ModelFact'
    :param rule_dict: A rule from the rule dict.
        E.g. a return value of `_parse_row`.
    :type rule_dict: dict
    :return: True if the fact matched the rule else False.
    :type: bool
    """
    fact_matches = False
    artifacts = get_artifact_lists(fact, rule_dict)
    for fact_artifact in artifacts:
        if rule_dict['relation'] == 'Contains':
            fact_matches = contains(fact_artifact, rule_dict['item_check'])
        elif rule_dict['relation'] == 'Contains_insensitive':
            fact_matches = contains_insensitive(
                fact_artifact, rule_dict['item_check']
            )
        elif rule_dict['relation'] == 'Equals':
            fact_matches = equals(fact_artifact, rule_dict['item_check'])
        elif rule_dict['relation'] == 'Has_member':
            fact_matches = equals(
                fact_artifact, rule_dict['item_check'].split('|')[1]
            )
        if fact_matches:
            # if fact matches rule condition escape loop,
            # otherwise continue checking
            break

    if rule_dict['negation'] == 'Not':
        fact_matches = not fact_matches

    if rule_dict['additional_conditions'] is None:
        return fact_matches
    else:
        return (
            fact_matches and
            check_rule(fact, rule_dict['additional_conditions'])
        )

# =====================Relationship checks=============================


def contains(fact_part, dict_check):
    """
    Check if the fact_part contains the dict_check item.

    :param fact_part: An arelle model object value pulled off of the current
        ModelFact being tested.
    :type fact_part: :class:'~arelle.InstanceModelObject.ModelFact'
    :param dict_check: Value from the current rule dict to be compared against.
    :type dict_check: dict
    :return: True if the dict_check is contained in fact_part else False.
    :rtype: bool
    """
    return dict_check in str(fact_part)


def contains_insensitive(fact_part, dict_check):
    """
    Check if the fact_part contains the dict_check item, ignoring case.

    :param fact_part: An arelle model object value pulled off of the current
        ModelFact being tested.
    :type fact_part: :class:'~arelle.InstanceModelObject.ModelFact'
    :param dict_check: Value from the current rule dict to be compared against.
    :type dict_check: dict
    :return: True if the dict_check is contained in fact_part else False.
    :rtype: bool
    """
    return dict_check.lower() in str(fact_part).lower()


def equals(fact_part, dict_check):
    """
    Check if the fact_part equals the dict_check item.

    :param fact_part: A string or number representing some aspect of a fact.
    :type fact_part: str or int
    :param dict_check: Value from the current rule dict to be compared against.
    :type dict_check: str
    :returns: True if dict_check == fact_part or
        if they are numbers and are equal, else False
    :rtype: bool
    """
    if dict_check == str(fact_part):
        return True
    try:
        coerced_dict_value = float(dict_check)
        return coerced_dict_value == fact_part
    except ValueError:
        return False


# =============================Find Artifacts ==============================


def get_artifact_lists(fact, rule_dict):
    """
    Given a ModelFact instance and an "artifact_type" key derived from the
    negative_numbers.csv file, lookup the corresponding field or object off of
    the given ModelFact and return its value.

    :param fact: An arelle ModelFact instance.
    :type fact: :class:'~arelle.InstanceModelFact'
    :param rule_dict: A dictionary with the rule information
    :type rule_dict: dict
    :return: An iterable of the arelle model object values pulled of the
        ModelFact corresponding to the passed in artifact_type.
    :rtype: iterable
    """
    artifacts = []
    artifact_type = rule_dict['artifact']

    if artifact_type == "Member":
        artifacts = facts.member_qnames(fact)
    if artifact_type == "Axis":
        if '|' in rule_dict['item_check']:
            artifacts = facts.member_qnames(
                fact, axis_filter=rule_dict['item_check'].split('|')[0]
            )
        else:
            artifacts = facts.axis_qnames(fact)

    return artifacts

# =============================Deal with CSV ================================


def concept_map_from_csv(neg_num_file):
    """
    Returns a map of {qname: id} of the concepts to test for the blacklist

    :param neg_num_file: The .csv file containing the list of concepts for
        this rule.
    :type neg_num_file: file
    :return: A map of {qname: id}.
    :rtype: dict
    """
    with open(neg_num_file, 'rt') as f:
        reader = csv.reader(f)
        return {row[1]: row[0] for row in reader}


def get_rules_from_csv(exclusion_file):
    """
    Return a list of rules for blacklist exclusions

    :return: a list representing the data from the negative_numbers.csv file.
    :rtype: list
    """
    blacklist_exclusion_rules = list()
    with open(exclusion_file, 'rt') as f:
        reader = csv.reader(f)
        next(reader)  # skip header
        for row in reader:
            rule = _parse_row(row[1:])
            if row[0] == 'BLE':
                blacklist_exclusion_rules.append(rule)
        return blacklist_exclusion_rules


def _parse_row(row):
    """
    Recursively move through a CSV row (ignoring the list inclusion values),
    slicing off the four relevant indexes each time.

    :param row: ([str, ..., str]) A list of strings, representing a row from
    the negative_numbers.csv file, with the first value removed
    :type row: list [str]
    :return: A dictionary representing the data from the CSV row.
    :rtype: dict
    """
    return {
        'artifact': row[0],
        'negation': row[1],
        'relation': row[2],
        'item_check': row[3],
        'additional_conditions': _parse_row(row[4:]) if len(row) > 4 else None
    }



# (c) Copyright 2015 - 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
from .util import facts, messages
import csv
import os


_CODE_NAME = 'DQC.US.0009'
_RULE_VERSION = '1.0.0'


def fact_gt_other_fact(val, *args, **kwargs):
    """
    Checks that the various fact pairs that are defined in the configuration
    are of an appropriate relative value (lesser or greater than one or the
    other, as per the config).

    :param val: val to to throw error from
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: No explicit return, but throws an error for things returned by
        _compare_facts
    :rtype: None
    """
    for index, lesser, greater in _read_csv():
        for fact_group in _compare_facts(lesser, greater, val):
            lesser_fact = fact_group[lesser]
            greater_fact = fact_group[greater]
            val.modelXbrl.error(
                '{base_code}.{index_id}'.format(
                    base_code=_CODE_NAME,
                    index_id=index
                ),
                messages.get_message(_CODE_NAME, index),
                modelObject=[lesser_fact, greater_fact],
                lesser_label=lesser_fact.concept.label(),
                greater_label=greater_fact.concept.label(),
                ruleVersion=_RULE_VERSION
            )


def _read_csv():
    """
    Reads in information from dqc_us_config.csv and returns
    it in a tuple

    :return: tuple of info read in from dqc_us_config.csv
    :rtype: tuple (str, str, str)
    """
    path = os.path.join(
        os.path.dirname(__file__),
        'resources/DQC_US_0009/dqc_us_9_config.csv'
    )
    with open(path, 'rt') as f:
        reader = csv.reader(f)
        next(reader)  # skip header
        for row in reader:
            yield row[0], row[1], row[2]


def _compare_facts(lesser, greater, val):
    """
    Compares the facts that are passed in based on the comparison data.
    Returns a list of facts that don't match up, if any.

    :param lesser: A dictionary with a 'lesser' us-gaap concept name
    :type lesser: dict
    :param greater: A dictionary 'greater' us-gaap concept name.
    :type greater: dict
    :param val: The validation information which includes a modelXbrl object.
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: A list of the fact pairs
    :rtype: list [tuple]
    """
    fact_dict = {
        lesser: facts.lookup_gaap_facts(lesser, val.modelXbrl),
        greater: facts.lookup_gaap_facts(greater, val.modelXbrl)
    }
    mapped_fact_groups = facts.prepare_facts_for_calculation(fact_dict)
    results = []
    for group in mapped_fact_groups:
        vals = facts.scale_values([group[greater], group[lesser]])
        greater_val = vals[0]
        lesser_val = vals[1]
        if greater_val < lesser_val:
            results.append(group)
    return results


__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': (
        'Checks pairs of facts to ensure '
        'that one is greater than the other.'
    ),
    # Mount points
    'Validate.XBRL.Finally': fact_gt_other_fact,
}



# (c) Copyright 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
import os

from .util import messages, neg_num
from .util import facts as facts_util


_CODE_NAME = 'DQC.US.0014'
_RULE_VERSION = '2.0.0'
_DEFAULT_CONCEPTS_FILE = os.path.join(
    os.path.dirname(__file__),
    'resources',
    'DQC_US_0014',
    'dqc_14_concepts.csv'
)


def run_negative_numbers_no_dimensions(val, *args, **kwargs):
    """
    Run the list of facts against our negative number check for facts with
    no dimensions and add errors for the hits in the various lists.

    :param val: The validation object which carries the validation information,
        including the ModelXBRL
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: No direct return but throws errors for facts matching
        the blacklist.
    :rtype: None
    """
    blacklist_dict = neg_num.concept_map_from_csv(_DEFAULT_CONCEPTS_FILE)
    blacklist_facts = filter_negative_number_no_dimensions_facts(
        val, blacklist_dict.keys()
    )
    for fact in blacklist_facts:
        index_key = blacklist_dict[fact.qname.localName]
        val.modelXbrl.error(
            '{base_key}.{extension_key}'.format(
                base_key=_CODE_NAME, extension_key=index_key
            ),
            messages.get_message(_CODE_NAME, str(index_key)),
            concept=fact.concept.label(),
            modelObject=fact,
            ruleVersion=_RULE_VERSION
        )


def filter_negative_number_no_dimensions_facts(val, blacklist_concepts):
    """
    Checks the numeric negative value dimensionless facts in the provided
    ModelXBRL instance against the rule dictionary and returns those which
    meet the conditions of the blacklist.

    :param val: The validation object which carries the validation information,
        including the ModelXBRL
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param blacklist_concepts: An iterable of the blacklist concepts we should
        be testing against.
    :type blacklist_concepts: list [str]
    :return: Returns a list of facts that fall in the blacklist
    :rtype: list [:class:'~arelle.ModelInstanceObject.ModelFact']
    """
    bad_blacklist = []

    numeric_facts = facts_util.grab_numeric_facts(list(val.modelXbrl.facts))
    # other filters before running negative numbers check
    # numeric_facts has already checked if fact.value can be made into a number
    facts_to_check = [
        fact for fact in numeric_facts
        if fact.xValue < 0 and
        fact.concept is not None and
        fact.concept.type is not None and
        # facts with numerical values less than 0 (negative) and contexts
        fact.context is not None and
        fact.context.segDimValues is not None and
        # check that the fact does not have dimensions
        len(fact.context.segDimValues) == 0
    ]

    # identify facts which should be reported as included in the list
    for fact in facts_to_check:
        if fact.qname.localName in blacklist_concepts:
            bad_blacklist.append(fact)

    return bad_blacklist


__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': (
        'Checks that the values for a given list of elements with no '
        'dimensions are negative'
    ),
    # Mount points
    'Validate.XBRL.Finally': run_negative_numbers_no_dimensions,
}



# (c) Copyright 2015 - 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
import os
import csv
from .util import messages, neg_num
from .util import facts as facts_util
from arelle.XmlValidate import VALID

_CODE_NAME = 'DQC.US.0015'
_RULE_VERSION = '2.0.0'
_DEFAULT_CONCEPTS_FILE = os.path.join(
    os.path.dirname(__file__),
    'resources',
    'DQC_US_0015',
    'dqc_15_concepts.csv'
)
_DEFAULT_EXCLUSIONS_FILE = os.path.join(
    os.path.dirname(__file__),
    'resources',
    'DQC_US_0015',
    'dqc_15_exclusion_rules.csv'
)


def run_negative_numbers(val, *args, **kwargs):
    """
    Run the list of facts against our negative number checks and add errors for
    the hits in the various lists.

    :param val: The validation object which carries the validation information,
        including the ModelXBRL
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: Nore direct return, but throws errors for facts matching the
        blacklist
    :rtype: None
    """
    # filter down to numeric facts
    blacklist_dict = neg_num.concept_map_from_csv(_DEFAULT_CONCEPTS_FILE)
    blacklist_facts = filter_negative_number_facts(val, blacklist_dict.keys())
    for fact in blacklist_facts:
        index_key = blacklist_dict[fact.qname.localName]
        val.modelXbrl.error(
            '{base_key}.{extension_key}'.format(
                base_key=_CODE_NAME, extension_key=index_key
            ),
            messages.get_message(_CODE_NAME, str(index_key)),
            concept=fact.concept.label(), modelObject=fact,
            ruleVersion=_RULE_VERSION
        )


def filter_negative_number_facts(val, blacklist_concepts):
    """
    Check the numeric negative value facts in the provided ModelXBRL instance
    against the rule dictionary, and return those which meet the conditions of
    the black list and aren't excluded.

    :param val: val whose modelXbrl provides the facts to check
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param blacklist_concepts: An iterable of the blacklist concepts we should
        be testing against.
    :type blacklist_concepts: list [str]
    :return: Return list of the facts falling into the blacklist.
    :rtype: list [:class:'~arelle.ModelInstanceObject.ModelFact']
    """
    blacklist_exclusion_rules = neg_num.get_rules_from_csv(
        _DEFAULT_EXCLUSIONS_FILE
    )
    bad_blacklist = []

    numeric_facts = facts_util.grab_numeric_facts(list(val.modelXbrl.facts))
    # other filters before running negative numbers check
    # numeric_facts has already checked if fact.value can be made into a number
    facts_to_check = [
        f for f in numeric_facts if f.xValid >= VALID and f.xValue < 0 and
        f.concept.type is not None and
        # facts with numerical values less than 0 and contexts and
        f.context is not None and
        # check xsd type of the concept
        f.isNumeric
    ]

    # identify facts which should be reported as included in the list
    for fact in facts_to_check:
        if neg_num.check_rules(fact, blacklist_exclusion_rules):
            continue  # cannot be black
        if ((fact.qname.localName in blacklist_concepts and
             fact.qname.namespaceURI in
             val.disclosureSystem.standardTaxonomiesDict)):
            bad_blacklist.append(fact)

    return bad_blacklist


def check_rules(fact, rule_dicts):
    for rule_dict in rule_dicts:
        if check_rule(fact, rule_dict):
            return True
    return False


def check_rule(fact, rule_dict):
    """
    Check if the input fact meets the conditions of the passed in rule_dict.

    :param fact: An arelle ModelFact instance.
    :type fact: :class:'~arelle.ModelInstanceObject.ModelFact'
    :param rule_dict: A rule from the rule dict.
        E.g. a return value of `_parse_row`.
    :type rule_dict: dict
    :return: True if the fact matched the rule else False.
    :type: bool
    """
    fact_matches = False
    artifacts = get_artifact_lists(fact, rule_dict)
    for fact_artifact in artifacts:
        if rule_dict['relation'] == 'Contains':
            fact_matches = contains(fact_artifact, rule_dict['item_check'])
        elif rule_dict['relation'] == 'Contains_insensitive':
            fact_matches = contains_ignore_case(
                fact_artifact, rule_dict['item_check']
            )
        elif rule_dict['relation'] == 'Equals':
            fact_matches = equals(fact_artifact, rule_dict['item_check'])
        elif rule_dict['relation'] == 'Has_member':
            fact_matches = equals(
                fact_artifact, rule_dict['item_check'].split('|')[1]
            )
        if fact_matches:
            # if fact matches rule condition escape loop,
            # otherwise continue checking
            break

    if rule_dict['negation'] == 'Not':
        fact_matches = not fact_matches

    if rule_dict['additional_conditions'] is None:
        return fact_matches
    else:
        return (
            fact_matches and
            check_rule(fact, rule_dict['additional_conditions'])
        )

# =====================Relationship checks=============================


def contains(fact_part, dict_check):
    """
    Check if the fact_part contains the dict_check item.

    :param fact_part: An arelle model object value pulled off of the current
        ModelFact being tested.
    :type fact_part: :class:'~arelle.InstanceModelObject.ModelFact'
    :param dict_check: Value from the current rule dict to be compared against.
    :type dict_check: dict
    :return: True if the dict_check is contained in fact_part else False.
    :rtype: bool
    """
    return dict_check in str(fact_part)


def contains_ignore_case(fact_part, dict_check):
    """
    Check if the fact_part contains the dict_check item, ignoring case.

    :param fact_part: An arelle model object value pulled off of the current
        ModelFact being tested.
    :type fact_part: :class:'~arelle.InstanceModelObject.ModelFact'
    :param dict_check: Value from the current rule dict to be compared against.
    :type dict_check: dict
    :return: True if the dict_check is contained in fact_part else False.
    :rtype: bool
    """
    return dict_check.lower() in str(fact_part).lower()


def equals(fact_part, dict_check):
    """
    Check if the fact_part equals the dict_check item.

    :param fact_part: A string or number representing some aspect of a fact.
    :type fact_part: str or int
    :param dict_check: Value from the current rule dict to be compared against.
    :type dict_check: str
    :returns: True if dict_check == fact_part or
        if they are numbers and are equal, else False
    :rtype: bool
    """
    if dict_check == str(fact_part):
        return True
    try:
        coerced_dict_value = float(dict_check)
        return coerced_dict_value == fact_part
    except ValueError:
        return False


# =============================Find Artifacts ==============================


def get_artifact_lists(fact, rule_dict):
    """
    Given a ModelFact instance and an "artifact_type" key derived from the
    negative_numbers.csv file, lookup the corresponding field or object off of
    the given ModelFact and return its value.

    :param fact: An arelle ModelFact instance.
    :type fact: :class:'~arelle.InstanceModelFact'
    :param rule_dict: A dictionary with the rule information
    :type rule_dict: dict
    :return: An iterable of the arelle model object values pulled of the
        ModelFact corresponding to the passed in artifact_type.
    :rtype: iterable
    """
    artifacts = []
    artifact_type = rule_dict['artifact']

    if artifact_type == "Member":
        artifacts = facts_util.member_qnames(fact)
    if artifact_type == "Axis":
        if '|' in rule_dict['item_check']:
            artifacts = facts_util.member_qnames(
                fact, axis_filter=rule_dict['item_check'].split('|')[0]
            )
        else:
            artifacts = facts_util.axis_qnames(fact)

    return artifacts

# =============================Deal with CSV ================================


def get_rules_from_csv():
    """
    Return a list of rules for blacklist exclusions

    :return: a list representing the data from the negative_numbers.csv file.
    :rtype: list
    """
    blacklist_exclusion_rules = list()
    with open(_DEFAULT_EXCLUSIONS_FILE, 'rt') as f:
        reader = csv.reader(f)
        next(reader)  # skip header
        for row in reader:
            rule = _parse_row(row[1:])
            if row[0] == 'BLE':
                blacklist_exclusion_rules.append(rule)
        return blacklist_exclusion_rules


def _parse_row(row):
    """
    Recursively move through a CSV row (ignoring the list inclusion values),
    slicing off the four relevant indexes each time.

    :param row: ([str, ..., str]) A list of strings, representing a row from
    the negative_numbers.csv file, with the first value removed
    :type row: list [str]
    :return: A dictionary representing the data from the CSV row.
    :rtype: dict
    """
    return {
        'artifact': row[0],
        'negation': row[1],
        'relation': row[2],
        'item_check': row[3],
        'additional_conditions': _parse_row(row[4:]) if len(row) > 4 else None
    }


__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': (
        'Checks all of the specified types and concepts for their '
        'date ranges to verify the ranges are within expected '
        'parameters for the fiscal periods'
    ),
    # Mount points
    'Validate.XBRL.Finally': run_negative_numbers,
}



# (c) Copyright 2015 - 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
import json
import os
from arelle.ModelDtsObject import ModelConcept
from .util import facts, messages
import itertools
from collections import defaultdict

_CODE_NAME = 'DQC.US.0001'
_RULE_VERSION = '2.0.0'
_DQC_01_AXIS_FILE = os.path.join(
    os.path.dirname(__file__),
    'resources',
    'DQC_US_0001',
    'dqc_0001.json'
)
_PARENT_CHILD_ARCROLE = 'http://www.xbrl.org/2003/arcrole/parent-child'
_ADDITIONAL_AXES_KEY = 'additional_axes'
_EXCLUDED_AXES_KEY = 'excluded_axes'
_DEFINED_MEMBERS_KEY = 'defined_members'
_ADDITIONAL_MEMBERS_KEY = 'additional_members'
_EXTENSIONS_KEY = 'extensions'
_RULE_INDEX_KEY = 'rule_index'
_UGT_FACT_KEY = 'ugt_fact'
_NO_FACT_KEY = 'no_fact'
_EXT_FACT_KEY = 'ext_fact'


def run_checks(val, *args, **kwargs):
    """
    Entrypoint for the rule.  Load the config, search for instances of
    each axis of interest, and call validations on them.

    :param val: val from which to gather end dates
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: No direct return
    :rtype: None
    """
    checked_axes = defaultdict(list)
    config = _load_config(_DQC_01_AXIS_FILE)
    for axis_key, axis_config in config.items():
        for role in val.modelXbrl.roleTypes:
            relset = val.modelXbrl.relationshipSet(
                _PARENT_CHILD_ARCROLE, linkrole=role
            )

            def filter_func(model_object):
                return (
                    _is_concept(model_object) and
                    model_object.qname.localName == axis_key
                )

            for axis in filter(filter_func, relset.fromModelObjects()):
                _run_axis_checks(
                    axis,
                    axis_key,
                    axis_config,
                    relset,
                    val,
                    role,
                    checked_axes
                )


def _run_axis_checks(axis, axis_key, axis_config, relset, val, role,
                     checked_axes):
    """
    Run the axis checks for a given axis, config dict,
    and set of children.

    :param axis: The axis to check
    :type axis: :class:'~arelle.ModelDTSObject.ModelConcept'
    :param axis_key: The axis name to check
    :type axis_key: str
    :param axis_config: The axis-specific config.
    :type axis_config: dict
    :param relset: The relationshipSet for the axis.
    :type relset: set
    :param val: val from which to gather end dates
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param role: The role for the relationship
    :type role: str
    :param checked_axes: Dictionary of already-fired axis_key: axis/member.
    :type checked_axes: dict
    :return: No direct return
    :rtype: None
    """
    _run_member_checks(
        axis,
        axis_key,
        axis_config,
        relset,
        val,
        role,
        checked_axes
    )
    _run_extension_checks(
        axis,
        axis_key,
        axis_config,
        relset,
        val,
        role,
        checked_axes
    )


def _run_member_checks(axis, axis_key, axis_config, relset, val, role,
                       checked_axes):
    """
    Run the checks on included and excluded members and companion axes.
    Extensions are not checked.  Error as appropriate.

    :param axis: The axis to check
    :type axis: :class:'~arelle.ModelDTSObject.ModelConcept'
    :param axis_key: The axis name to check
    :type axis_key: str
    :param axis_config: The axis-specific config.
    :type axis_config: dict
    :param relset: The relationshipSet for the axis.
    :type relset: set
    :param val: val from which to gather end dates
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param role: The role for the relationship
    :type role: str
    :param checked_axes: Dictionary of already-fired axis_key: axis/member.
    :type checked_axes: dict
    :return: No direct return
    :rtype: None
    """
    additional_axes = axis_config.get(_ADDITIONAL_AXES_KEY, {})
    excluded_axes = axis_config.get(_EXCLUDED_AXES_KEY, {})
    allowed_children = (
        axis_config.get(_DEFINED_MEMBERS_KEY, []) +
        axis_config.get(_ADDITIONAL_MEMBERS_KEY, [])
    )
    disallowed_children = []
    disallowed_children.extend(
        itertools.chain.from_iterable(
            member_list for member_list in excluded_axes.values()
        )
    )
    allowed_children.extend(
        itertools.chain.from_iterable(
            member_list for member_list in additional_axes.values()
        )
    )
    if len(disallowed_children) > 0:
        # Blacklisted axes check - Can only check blacklist (excluded)
        # or whitelist (included) axes.
        # Default to blacklist if both are present.
        for child in _all_members_under(axis, relset):
            if ((not _is_extension(child, val) and
                    child.qname.localName in disallowed_children)):
                axis_mem_pair = (axis.qname, child.qname)
                if axis_mem_pair in checked_axes[axis_key]:
                    continue
                fact_list = facts.axis_member_fact(
                    axis.qname.localName,
                    child.qname.localName,
                    val.modelXbrl
                )

                for fact in fact_list:  # remove grouped messages
                    val.modelXbrl.error(
                        '{base_key}.{extension_key}'.format(
                            base_key=_CODE_NAME,
                            extension_key=axis_config[_RULE_INDEX_KEY]
                        ),
                        messages.get_message(_CODE_NAME, _UGT_FACT_KEY),
                        axis=axis.label(),
                        member=child.label(),
                        modelObject=fact,
                        ruleVersion=_RULE_VERSION
                    )
                checked_axes[axis_key].append(axis_mem_pair)
                if len(fact_list) == 0:
                    val.modelXbrl.error(
                        '{base_key}.{extension_key}'.format(
                            base_key=_CODE_NAME,
                            extension_key=axis_config[_RULE_INDEX_KEY]
                        ),
                        messages.get_message(_CODE_NAME, _NO_FACT_KEY),
                        axis=axis.label(),
                        member=child.label(),
                        group=role,
                        ruleVersion=_RULE_VERSION
                    )
    else:
        # Whitelisted axes are specified.
        for child in _all_members_under(axis, relset):
            if ((not _is_extension(child, val) and
                 child.qname.localName not in allowed_children)):
                axis_mem_pair = (axis.qname, child.qname)
                if axis_mem_pair in checked_axes[axis_key]:
                    continue
                fact_list = facts.axis_member_fact(
                    axis.qname.localName,
                    child.qname.localName,
                    val.modelXbrl
                )
                for fact in fact_list:  # remove for grouped messages
                    val.modelXbrl.error(
                        '{base_key}.{extension_key}'.format(
                            base_key=_CODE_NAME,
                            extension_key=axis_config[_RULE_INDEX_KEY]
                        ),
                        messages.get_message(_CODE_NAME, _UGT_FACT_KEY),
                        axis=axis.label(),
                        member=child.label(),
                        modelObject=fact,
                        ruleVersion=_RULE_VERSION,

                    )
                checked_axes[axis_key].append(axis_mem_pair)
                if len(fact_list) == 0:
                    val.modelXbrl.error(
                        '{base_key}.{extension_key}'.format(
                            base_key=_CODE_NAME,
                            extension_key=axis_config[_RULE_INDEX_KEY]
                        ),
                        messages.get_message(_CODE_NAME, _NO_FACT_KEY),
                        axis=axis.label(),
                        member=child.label(),
                        group=role,
                        ruleVersion=_RULE_VERSION
                    )


def _run_extension_checks(axis, axis_key, axis_config, relset, val, role,
                          checked_axes):
    """
    Check extension members under the given axis.

    :param axis: The axis to check
    :type axis: :class:'~arelle.ModelDTSObject.ModelConcept'
    :param axis_key: The axis name to check
    :type axis_key: str
    :param axis_config: The axis-specific config.
    :type axis_config: dict
    :param relset: The relationshipSet for the axis.
    :type relset: set
    :param val: val from which to gather end dates
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param role: The role for the relationship
    :type role: str
    :param checked_axes: Dictionary of already-fired axis_key: axis/member.
    :type checked_axes: dict
    :return: No direct return
    :rtype: None
    """
    allow_all = (
        len(axis_config[_EXTENSIONS_KEY]) > 0 and
        axis_config[_EXTENSIONS_KEY][0] == '*'
    )
    if not allow_all:
        allowed_extensions = axis_config[_EXTENSIONS_KEY]
        for child in _all_members_under(axis, relset):
            if _is_extension(child, val):
                if child.qname.localName not in allowed_extensions:
                    axis_mem_pair = (axis.qname, child.qname)
                    if axis_mem_pair in checked_axes[axis_key]:
                        continue
                    fact_list = facts.axis_member_fact(
                        axis.qname.localName,
                        child.qname.localName,
                        val.modelXbrl
                    )
                    for fact in fact_list:  # remove for grouped messages
                        val.modelXbrl.error(
                            '{base_key}.{extension_key}'.format(
                                base_key=_CODE_NAME,
                                extension_key=axis_config[_RULE_INDEX_KEY]
                            ),
                            messages.get_message(
                                _CODE_NAME, _EXT_FACT_KEY
                            ),
                            axis=axis.label(),
                            member=child.label(),
                            modelObject=fact,
                            ruleVersion=_RULE_VERSION,
                        )
                    checked_axes[axis_key].append(axis_mem_pair)
                    if len(fact_list) == 0:
                        val.modelXbrl.error(
                            '{base_key}.{extension_key}'.format(
                                base_key=_CODE_NAME,
                                extension_key=axis_config[_RULE_INDEX_KEY]
                            ),
                            messages.get_message(_CODE_NAME, _NO_FACT_KEY),
                            axis=axis.label(),
                            member=child.label(),
                            group=role,
                            ruleVersion=_RULE_VERSION
                        )


def _is_extension(concept, val):
    """
    Return True if concept is an extension.

    :param concept: The concept to check
    :type concept: :class:'~arelle.ModelDTSObject.ModelConcept'
    :param val: val from which to gather end dates
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: True if concept is an extension, else False
    :rtype: bool
    """
    return (
        concept.qname.namespaceURI not in
        val.disclosureSystem.standardTaxonomiesDict
    )


def _is_concept(concept):
    """
    This utility method should be used instead of None checks on
    arc.fromModelObject or arc.toModelObject.

    :param concept: The concept to check
    :type concept: :class:'~arelle.ModelDTSObject.ModelConcept'
    :return: True if it's a valid concept.  False if not.
    :rtype: bool
    """
    return (
        isinstance(concept, ModelConcept) and
        concept.qname
    )


def _all_members_under(axis, relset):
    """
    Returns a dictionary of concepts seen under the provided concept,
    in the given relset.

    Dictionary values are locators for the concepts: the `toLocator`
    from the arc where that concept was discovered.

    :param axis: The axis to check.
    :type axis: :class:'~arelle.ModelDTSObject.ModelConcept'
    :param relset: The relationshipSet.
    :type relset: set
    :return: a list of members under the axis.
    :rtype: [:class:'~arelle.ModelDTSObject.ModelConcept']
    """
    concepts = []
    arcs_to_check = []
    seen_arcs = set()
    for arc in relset.fromModelObject(axis):
        seen_arcs.add(arc)
        arcs_to_check.append(arc)
    while(arcs_to_check):
        cur_arc = arcs_to_check.pop()
        to_object = cur_arc.toModelObject
        if _is_concept(to_object) and not _is_domain(to_object):
            concepts.append(to_object)
        for arc in relset.fromModelObject(to_object):
            if arc not in seen_arcs:
                seen_arcs.add(arc)
                arcs_to_check.append(arc)
    return concepts


def _is_domain(concept):
    """
    Return True if the concept is a domain, else False

    :param concept: The concept to check.
    :type concept: :class:'~arelle.ModelDTSObject.ModelConcept'
    :return: True if the concept is a domain, else False.
    :rtype: bool
    """
    return (
        '[Domain]' in concept.label() or
        concept.qname.localName.endswith('Domain')
    )


def _load_config(axis_file):
    """
    Returns a map of axis/configs to test.

    :param axis_file: the file to open.
    :type axis_file: file
    :return: A map of the config file.
    :rtype: dict
    """

    with open(axis_file) as config:
        return json.load(config)

__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': 'Axis member checks.',
    # Mount points
    'Validate.XBRL.Finally': run_checks,
}
